/*
 *    ImageServlet
 *  Copyright (C) 2010 Mark Elam
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 *  For more information about the GNU Public license see:
 *  http://www.gnu.org/copyleft/gpl.html
 *
 *  ImageServlet is at:
 *  http://www.subtrack.org.uk
 *
 */
package uk.org.subtrack.imaging

import javax.servlet.http.{HttpServlet, HttpServletRequest, HttpServletResponse}
import scala.collection.mutable.{HashMap, SynchronizedMap, Map}
import java.awt.image.BufferedImage
import java.io.IOException
import java.lang.reflect.InvocationTargetException
import java.util.Properties
import javax.imageio.ImageIO
import java.io.File
import java.io.ByteArrayInputStream
import java.io.ByteArrayOutputStream
import java.io.ObjectInputStream
import java.io.ObjectOutputStream

class ImageServlet extends HttpServlet with Logging {
  //using tuple but could use
  //private class ImageInfo(val imageBean:ImageBean, val image:Image )
  private val imageMap = Map.empty[String, (ImageBean, ImagesDocument.Image)]
  private val mimeTypes = Map.empty[String, String]
  private val bufferedImageMap = new HashMap[String, BufferedImage]() with SynchronizedMap[String, BufferedImage]
  private val defaultFormat = "png"
  private val defaultMIMEType = "image/png"
  private val defaultPropertiesFile = "/imaging.properties"
  private val defaultImagesFile = "/images.xml"
  private val ImagesFile = "IMAGES_FILE"
  private val PropertiesFile = "PROPERTIES_FILE"

  override def init() {
    logger.debug("init called in ImageServlet")
    readProperties()
    readImagesXML()
    super.init()
  }

  private def getParameter(paramName: String, default: String): String = {
    //some implicit def to convert Enumerations to Map would remove null use
    val param = getInitParameter(paramName)
    if (param != null) param else default
  }

  private def readProperties() {
    val propertiesFile = getParameter(PropertiesFile, defaultPropertiesFile)
    logger.debug("read properties from " + propertiesFile)
    try {
      val props = new Properties()
      props.load(classOf[ImageServlet].getResourceAsStream(propertiesFile))
      val keys = props.keySet.toArray[String](new Array[String](props.size))
      for (key <- keys) {
        if (key.startsWith("imaging.format.")) {
          val format = key.substring(key.lastIndexOf(".") + 1)
          logger.debug("add to mimeTypes:" + format + ":" + props.get(key))
          mimeTypes += (format -> props.get(key).toString)
        }
      }
    } catch {
      case ex: IOException => logger.debug("no Imaging Properties file found, using defaults", ex)
    }
  }

  private def readImagesXML() {
    val imagesFile = getParameter(ImagesFile, defaultImagesFile)
    logger.debug("read images from " + imagesFile)
    try {
      val is = classOf[ImageServlet].getResourceAsStream(imagesFile)
      val imagesDocument = ImagesDocument.Images(is)
      for (image <- imagesDocument) {
        logger.debug(image.name + "," + image.classname)
        initiliseImage(image)
      }
    } catch {
      case ex: IOException => logger.fatal("Unable to initialise ImagingServlet", ex)
    }
  }

  private def splitImageName(fullImageName: String): Tuple3[String, String, String] = {
    //strip off type
    val splits = fullImageName split "[.]"
    //check there are 2 splits and assign accordingly
    if (splits.length > 1) {
      mimeTypes.get(splits(1)) match {
        case Some(mtype) => (splits(0), mtype, splits(1))
        case None => (splits(0), defaultMIMEType, defaultFormat)
      }
    }
    else {
      (fullImageName, defaultMIMEType, defaultFormat)
    }
  }

  override def doGet(req: HttpServletRequest, res: HttpServletResponse) {
    logger.debug("Path info" + req.getPathInfo())
    val out = res.getOutputStream() // binary output!
    val fullImageName = req.getPathInfo()
    val (imageName, mimeType, format) = splitImageName(fullImageName)
    logger.debug(String.format("imageName:%s format:%s mimeType:%s", imageName, format, mimeType))
    val imageInfo = imageMap.get(imageName)
    if (imageInfo != None) {
      //copy synchronized which produces local copy for use
      //do we need copy with immutable objects?(object not really immutable so yes!)
      val imageBean = copy(imageInfo.get._1).get
      val image = imageInfo.get._2
      //unique identifier for each image is image name plus parameters
      val fullQueryString = imageName + req.getQueryString()
      val bufferedImageOption = retrieveFromImageCache(fullQueryString)
      val bufferedImage = {
        if (bufferedImageOption == None) {
          logger.debug("Create new Image for " + fullQueryString)
          //if first request generate Image and store in cache
          setQueryStringParameters(imageBean, req)
          val size = image.size
          val baseImage = image.baseimage
          val drawImage = (image: BufferedImage, height: Int, width: Int) => {
            imageBean.drawGraphics(image.createGraphics(), height, width)
            image
          }
          val bufferedImage = {
            if (baseImage.length > 0) {
              val translatedPath = req.getPathTranslated()
              val path = translatedPath.substring(0, translatedPath.lastIndexOf(File.separatorChar))
              val imageFileName = path + req.getServletPath() + baseImage
              val bufferedImage = ImageIO.read(new File(imageFileName))
              drawImage(bufferedImage, bufferedImage.getHeight(), bufferedImage.getWidth())
            } else {
              val bufferedImage = new BufferedImage(size.width.intValue(), size.height.intValue(), BufferedImage.TYPE_INT_RGB)
              drawImage(bufferedImage, size.height.intValue(), size.width.intValue())
            }
          }
          storeInImageCache(fullQueryString, bufferedImage)
          bufferedImage
        } else {
          bufferedImageOption.get
        }
      }
      logger.debug("generate image as: ." + format + ":" + mimeType)
      ImageIO.write(bufferedImage, format, out)
      res.setContentType(mimeType)
    }
    // Set to expire far in the past.
    res.setHeader("Expires", "Sat, 6 May 1995 12:00:00 GMT")

    // Set standard HTTP/1.1 no-cache headers.
    res.setHeader("Cache-Control", "no-store, no-cache, must-revalidate")

    // Set IE extended HTTP/1.1 no-cache headers (use addHeader).
    res.addHeader("Cache-Control", "post-check=0, pre-check=0")

    // Set standard HTTP/1.0 no-cache header.
    res.setHeader("Pragma", "no-cache")
  }

  private def initiliseImage(image: ImagesDocument.Image) {
    try {
      val imageClass = Class.forName(image.classname)
      val imageObj = imageClass.newInstance()
      //get rid of casts etc.
      if (imageObj.isInstanceOf[ImageBean]) {
        val imageBean = imageObj.asInstanceOf[ImageBean]
        //add any properties to new Bean if possible        
        for (property <- image.properties) {
          initialiseProperty(imageBean, property)
        }
        //store in Map
        imageMap += (image.name -> (imageBean, image))
      }


    } catch {
      case ex: InstantiationException => logger.debug(ex)
      case ex: IllegalAccessException => logger.debug(ex)
      case ex: ClassNotFoundException => logger.debug("Class " + image.classname + " Not found", ex)
    }
  }

  private def initialiseProperty(imageBean: ImageBean, property: ImagesDocument.Property) {
    initialiseProperty(imageBean, property.name, property.value)
  }

  private def initialiseProperty(imageBean: ImageBean, propertyName: String, propertyValue: String) {
    def trace1() {
      logger.trace("No field " + propertyName + " found on [" + imageBean.getClass().getName() + "]")
    }
    //try field first...
    var set = false
    try {
      val field = imageBean.getClass().getField(propertyName)
      field.set(imageBean, propertyValue)
      set = true
    } catch {
      case ex: IllegalArgumentException => trace1
      case ex: IllegalAccessException => trace1
      case ex: NoSuchFieldException => trace1
      case ex: SecurityException => trace1
    }
    //try method otherwise
    if (!set) {
      val methodName = "set" + Character.toUpperCase(propertyName.charAt(0)) + propertyName.substring(1)
      def trace2() {
        logger.trace("No Method " + methodName + " available on [" + imageBean.getClass().getName() + "]")
      }
      try {
        val method = imageBean.getClass().getMethod(methodName, Array(classOf[String]): _*)
        method.invoke(imageBean, propertyValue)
      } catch {
        case ex: IllegalAccessException => trace2
        case ex: IllegalArgumentException => trace2
        case ex: InvocationTargetException => trace2
        case ex: NoSuchMethodException => trace2
        case ex: SecurityException => trace2
      }
    }
  }

  private def setQueryStringParameters(imageBean: ImageBean, req: HttpServletRequest) {
    val parameters = req.getParameterMap()
    val properties = parameters.keySet.toArray[String](new Array[String](parameters.size))
    for (property <- properties) {
      val parameter = parameters.get(property).asInstanceOf[Array[String]](0)
      logger.debug("Set parameter:" + property + " to " + parameter)
      initialiseProperty(imageBean, property, parameter)
    }

  }

  private def retrieveFromImageCache(fullQueryString: String): Option[BufferedImage] = {
    logger.debug("Retrieve previous image for " + fullQueryString)
    if (bufferedImageMap.contains(fullQueryString))
      Some(bufferedImageMap(fullQueryString))
    else
      None
  }

  private def storeInImageCache(qString: String, bufferedImage: BufferedImage) {
    logger.debug("Store image for " + qString)
    bufferedImageMap(qString) = bufferedImage
  }

  //todo:find what should be used here
  //  private def copy(orig:ImageBean ) :ImageBean ={
  //    orig
  //  }
  //was synchronised
  private def copy(orig: ImageBean): Option[ImageBean] = {
    try {
      // Write the object out to a byte array
      val bos = new ByteArrayOutputStream()
      val out = new ObjectOutputStream(bos)
      out.writeObject(orig)
      out.flush()
      out.close()

      // Make an input stream from the byte array and read
      // a copy of the object back in.
      val in = new ObjectInputStream(
        new ByteArrayInputStream(bos.toByteArray()))
      Some(in.readObject().asInstanceOf[ImageBean])
    } catch {
      case e: IOException => logger.fatal("Cannot copy ImageBean", e)
      None
      case e: ClassNotFoundException => logger.fatal("Cannot copy ImageBean", e)
      None
    }
  }
}
